(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: ShadowVolume.cpp
//
// Desc: Example code showing how to use stencil buffers to implement shadow
//       volumes.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit ShadowVolumeUnit;

interface

uses
  Windows, Messages, SysUtils, Resource, 
  Direct3D9, D3DX9, D3DApp, D3DFile, D3DFont, D3DUtil, DXUtil;

//-----------------------------------------------------------------------------
// External definitions and prototypes
//-----------------------------------------------------------------------------
function FtoDW(f: Single): DWORD;

type
  TVertex = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
    tu, tv: Single;
  end;
  PVertexArray = ^TVertexArray;
  TVertexArray = array [0..0] of TVertex;

const
  TVertex_FVF = D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX1;

type
  TShadowVertex = packed record
    p: TD3DXVector4;
    color: TD3DColor;
  end;
  PShadowVertexArray = ^TShadowVertexArray;
  TShadowVertexArray = array [0..MaxInt div SizeOf(TShadowVertex) - 1] of TShadowVertex;

const
  TShadowVertex_FVF = D3DFVF_XYZRHW or D3DFVF_DIFFUSE;

type
  //-----------------------------------------------------------------------------
  // Name: struct ShadowVolume
  // Desc: A shadow volume object
  //-----------------------------------------------------------------------------
  TShadowVolume = class
  private
    m_pVertices: array [0..32000-1] of TD3DXVector3; // Vertex data for rendering shadow volume
    m_dwNumVertices: DWORD;
  public
    procedure Reset; { m_dwNumVertices = 0L; }
    function BuildFromMesh(pMesh: ID3DXMesh; vLight: TD3DXVector3): HResult;
    function Render(pd3dDevice: IDirect3DDevice9): HResult;
  end;

  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont:            ID3DXFont;
    m_Camera:           CModelViewerCamera;
    m_bShowHelp:        Boolean;

    m_pAirplane:        CD3DMesh;
    m_pTerrainObject:   CD3DMesh;

    m_pShadowVolume:    TShadowVolume;

    m_matObjectMatrix:  TD3DXMatrixA16;
    m_matTerrainMatrix: TD3DXMatrixA16;

    m_pBigSquareVB:     IDirect3DVertexBuffer9;

    function DrawShadow: HResult;
    function RenderShadow: HResult;
    procedure RenderText;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; overload; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
  public
    constructor Create;
    destructor Destroy; override;

    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
 end;

implementation

function FtoDW(f: Single): DWORD;
begin
  Result:= PDWord(@f)^;
end;

{ TShadowVolume }

//-----------------------------------------------------------------------------
// Name:
// Desc:
//-----------------------------------------------------------------------------
function TShadowVolume.Render(pd3dDevice: IDirect3DDevice9): HResult;
begin
  pd3dDevice.SetFVF(D3DFVF_XYZ);

  Result:= pd3dDevice.DrawPrimitiveUP(D3DPT_TRIANGLELIST, m_dwNumVertices div 3,
                                      m_pVertices, SizeOf(TD3DXVector3));
end;

//-----------------------------------------------------------------------------
// Name: AddEdge()
// Desc: Adds an edge to a list of silohuette edges of a shadow volume.
//-----------------------------------------------------------------------------
procedure AddEdge(var pEdges: array of Word; var dwNumEdges: DWord; v0, v1: Word);
var
  i: Integer;
begin
  // Remove interior edges (which appear in the list twice)
  for i:= 0 to (dwNumEdges - 1) do
  begin
    if ((pEdges[2*i+0] = v0) and (pEdges[2*i+1] = v1)) or
       ((pEdges[2*i+0] = v1) and (pEdges[2*i+1] = v0)) then
    begin
      if (dwNumEdges > 1) then
      begin
        pEdges[2*i+0] := pEdges[2*(dwNumEdges-1)+0];
        pEdges[2*i+1] := pEdges[2*(dwNumEdges-1)+1];
      end;
      Dec(dwNumEdges);
      Exit;
    end;
  end;

  pEdges[2*dwNumEdges+0] := v0;
  pEdges[2*dwNumEdges+1] := v1;
  Inc(dwNumEdges);
end;




//-----------------------------------------------------------------------------
// Name: BuildFromMesh()
// Desc: Takes a mesh as input, and uses it to build a shadowvolume. The
//       technique used considers each triangle of the mesh, and adds it's
//       edges to a temporary list. The edge list is maintained, such that
//       only silohuette edges are kept. Finally, the silohuette edges are
//       extruded to make the shadow volume vertex list.
//-----------------------------------------------------------------------------
function TShadowVolume.BuildFromMesh(pMesh: ID3DXMesh;
  vLight: TD3DXVector3): HResult;
type
  // Note: the "TMeshVertex" format depends on the FVF of the mesh
  PMeshVertex = ^TMeshVertex;
  TMeshVertex = packed record
    p, n: TD3DXVector3;
    tu, tv: Single;
  end;

  PMeshVertexArray = ^TMeshVertexArray;
  TMeshVertexArray = array [0..0] of TMeshVertex;

var
  pVertices: PMeshVertexArray;
  pIndices: PWordArray;

  dwNumFaces: DWORD;

  pEdges: array of Word;
  dwNumEdges: DWord;

  i: Integer;

  wFace0, wFace1, wFace2: Word;
  v0, v1, v2, v3, v4: TD3DXVector3;
  v001: TD3DXVector3;

  vNormal: TD3DXVector3;
  vCross1: TD3DXVector3;
  vCross2: TD3DXVector3;
begin
  // Lock the geometry buffers
  pMesh.LockVertexBuffer(0, Pointer(pVertices));
  pMesh.LockIndexBuffer(0, Pointer(pIndices));
  dwNumFaces := pMesh.GetNumFaces;

  // Allocate a temporary edge list
  try
    SetLength(pEdges, dwNumFaces*6);
  except
    on EOutOfMemory do
    begin
      pMesh.UnlockVertexBuffer;
      pMesh.UnlockIndexBuffer;
      Result:= E_OUTOFMEMORY;
      Exit;
    end;
  end;
  dwNumEdges:= 0;

  // For each face
  for i:= 0 to (dwNumFaces - 1) do
  begin
    wFace0 := pIndices[3*i+0];
    wFace1 := pIndices[3*i+1];
    wFace2 := pIndices[3*i+2];

    v0 := pVertices[wFace0].p;
    v1 := pVertices[wFace1].p;
    v2 := pVertices[wFace2].p;

    // Transform vertices or transform light?
    D3DXVec3Subtract(vCross1, v2, v1);
    D3DXVec3Subtract(vCross2, v1, v0);
    D3DXVec3Cross(vNormal, vCross1, vCross2);

    if (D3DXVec3Dot(vNormal, vLight) >= 0.0) then
    begin
      AddEdge(pEdges, dwNumEdges, wFace0, wFace1);
      AddEdge(pEdges, dwNumEdges, wFace1, wFace2);
      AddEdge(pEdges, dwNumEdges, wFace2, wFace0);
    end;
  end;

  for i:= 0 to (dwNumEdges - 1) do
  begin
    v1 := pVertices[pEdges[2*i+0]].p;
    v2 := pVertices[pEdges[2*i+1]].p;
    D3DXVec3Scale(v001, vLight, 10);
    // D3DXVECTOR3 v3 = v1 - vLight*10;
    D3DXVec3Subtract(v3, v1, v001);
    // D3DXVECTOR3 v4 = v2 - vLight*10;
    D3DXVec3Subtract(v4, v2, v001);

    // Add a quad (two triangles) to the vertex list
    m_pVertices[m_dwNumVertices] := v1; Inc(m_dwNumVertices);
    m_pVertices[m_dwNumVertices] := v2; Inc(m_dwNumVertices);
    m_pVertices[m_dwNumVertices] := v3; Inc(m_dwNumVertices);

    m_pVertices[m_dwNumVertices] := v2; Inc(m_dwNumVertices);
    m_pVertices[m_dwNumVertices] := v4; Inc(m_dwNumVertices);
    m_pVertices[m_dwNumVertices] := v3; Inc(m_dwNumVertices);
  end;
  // Delete the temporary edge list
  SetLength(pEdges, 0);

  // Unlock the geometry buffers
  pMesh.UnlockVertexBuffer;
  pMesh.UnlockIndexBuffer;

  Result := S_OK;
end;

procedure TShadowVolume.Reset;
begin
  m_dwNumVertices:= 0;
end;

{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;
  m_Camera := CModelViewerCamera.Create;

  m_strWindowTitle    := 'ShadowVolume: RealTime Shadows Using The StencilBuffer';
  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_d3dEnumeration.AppMinDepthBits := 16;
  m_d3dEnumeration.AppMinStencilBits := 4;
  m_bShowCursorWhenFullscreen := True;

  m_pFont           := nil;
  m_bShowHelp       := False;
  m_pAirplane       := CD3DMesh.Create;
  m_pTerrainObject  := CD3DMesh.Create;
  m_pShadowVolume   := nil;
  m_pBigSquareVB    := nil;
end;

destructor CMyD3DApplication.Destroy;
begin
  m_Camera.Free;
  inherited;
end;


//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  // Construct a shadow volume object;
  m_pShadowVolume:= TShadowVolume.Create;

  // Set cursor to indicate that user can move the object with the mouse
{$IFDEF WIN64}
  SetClassLongPtr(m_hWnd, GCLP_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ELSE}
  SetClassLong(m_hWnd, GCL_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ENDIF}
  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  hDC: Windows.HDC;
  nHeight: Integer;
  pVB: IDirect3DVertexBuffer9;
  pVertices: PVertexArray;
  dwNumVertices: DWORD;
  i: Integer;
begin
  // Initialize the font
  hDC := GetDC(0);
  nHeight := -MulDiv(9, GetDeviceCaps(hDC, LOGPIXELSY), 72);
  ReleaseDC(0, hDC);
  Result := D3DXCreateFont(m_pd3dDevice, nHeight, 0, FW_BOLD, 0, FALSE, DEFAULT_CHARSET,
              OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH or FF_DONTCARE,
              'Arial', m_pFont);
  if FAILED(Result) then Exit;

  Result := D3DAPPERR_MEDIANOTFOUND;

  // Load an object to cast the shadow
  if FAILED(m_pAirplane.Create_(m_pd3dDevice, PChar('airplane 2.x')))
  then Exit;

  // Load some terrain
  if FAILED(m_pTerrainObject.Create_(m_pd3dDevice, PChar('SeaFloor.x')))
  then Exit;

  // Set a reasonable vertex type
  m_pAirplane.SetFVF(m_pd3dDevice, TVertex_FVF);
  m_pTerrainObject.SetFVF(m_pd3dDevice, TVertex_FVF);

  // Tweak the terrain vertices
  begin
    dwNumVertices := m_pTerrainObject.GetSysMemMesh.GetNumVertices;

    // Lock the vertex buffer to access the terrain geometry
    m_pTerrainObject.GetSysMemMesh.GetVertexBuffer(pVB);
    pVB.Lock(0, 0, Pointer(pVertices), 0);

    // Add some more bumpiness to the terrain object
    for i:= 0 to (dwNumVertices - 1) do
    begin
      pVertices[i].p.y := pVertices[i].p.y + 1*(Random{rand/(FLOAT)RAND_MAX});
      pVertices[i].p.y := pVertices[i].p.y + 2*(Random{rand/(FLOAT)RAND_MAX});
      pVertices[i].p.y := pVertices[i].p.y + 1*(Random{rand/(FLOAT)RAND_MAX});
    end;

    // Release the vertex buffer
    pVB.Unlock;
    pVB:= nil;
  end;

  // Create a big square for rendering the stencilbuffer contents
  if FAILED(m_pd3dDevice.CreateVertexBuffer(4*SizeOf(TShadowVertex),
      D3DUSAGE_WRITEONLY, TShadowVertex_FVF, D3DPOOL_MANAGED, m_pBigSquareVB, nil)) then
  begin
    Result := E_FAIL;
    Exit;
  end;

  Result := S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  mtrl: TD3DMaterial9;

  vEyePt, vLookatPt: TD3DXVector3;
  vMin, vMax: TD3DXVector3;

  fAspect: Single;
  fFogStart, fFogEnd: Single;
  sx, sy: Single;
  v: PShadowVertexArray;
begin
  if Assigned(m_pFont) then m_pFont.OnResetDevice;

  // Initialize the vertex buffers for the file-based objects
  m_pAirplane.RestoreDeviceObjects(m_pd3dDevice);
  m_pTerrainObject.RestoreDeviceObjects(m_pd3dDevice);

  // Create and set up the shine materials w/ textures
  ZeroMemory(@mtrl, SizeOf(TD3DMaterial9));
  mtrl.Diffuse.r := 1.0; mtrl.Ambient.r := 1.0;
  mtrl.Diffuse.g := 1.0; mtrl.Ambient.g := 1.0;
  mtrl.Diffuse.b := 1.0; mtrl.Ambient.b := 1.0;
  mtrl.Diffuse.a := 1.0; mtrl.Ambient.a := 1.0;
  m_pd3dDevice.SetMaterial(mtrl);

  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, iFalse);

  // Set the transform matrices
  vEyePt    := D3DXVector3(0.0, 10.0, -20.0);
  vLookatPt := D3DXVector3(0.0,  0.0,   0.0);
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;

  // Setup the camera
  m_Camera.SetViewParams(vEyePt, vLookatPt);
  m_Camera.SetProjParams(D3DX_PI/4, fAspect, 1.0, 100.0);
  m_Camera.SetScalers(0.01, 50.0);
  m_Camera.SetEnablePositionMovement(True);
  m_Camera.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height, 0.9);
  m_Camera.SetRadius(15.0, 5.0, 40.0);
//  m_Camera.SetLimitPitch( false );
  vMin := D3DXVector3(-20,-5,-20);
  vMax := D3DXVector3(20,2,20);
  m_Camera.SetClipToBoundary(True, vMin, vMax);

  // Turn on fog
  fFogStart :=  30.0;
  fFogEnd   :=  80.0;
  m_pd3dDevice.SetRenderState(D3DRS_FOGENABLE,      iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_FOGCOLOR,       $ff00bfff);
  m_pd3dDevice.SetRenderState(D3DRS_FOGTABLEMODE,   D3DFOG_NONE);
  m_pd3dDevice.SetRenderState(D3DRS_FOGVERTEXMODE,  D3DFOG_LINEAR);
  m_pd3dDevice.SetRenderState(D3DRS_RANGEFOGENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_FOGSTART,       FtoDW(fFogStart));
  m_pd3dDevice.SetRenderState(D3DRS_FOGEND,         FtoDW(fFogEnd));

  m_pd3dDevice.LightEnable(0, True);
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,  $00303030);

  // Set the size of the big square shadow
  sx:= m_d3dsdBackBuffer.Width;
  sy:= m_d3dsdBackBuffer.Height;
  m_pBigSquareVB.Lock(0, 0, Pointer(v), 0);
  v[0].p := D3DXVector4( 0, sy, 0.0, 1.0);
  v[1].p := D3DXVector4( 0,  0, 0.0, 1.0);
  v[2].p := D3DXVector4(sx, sy, 0.0, 1.0);
  v[3].p := D3DXVector4(sx,  0, 0.0, 1.0);
  v[0].color := $7f000000;
  v[1].color := $7f000000;
  v[2].color := $7f000000;
  v[3].color := $7f000000;
  m_pBigSquareVB.Unlock;

  Result := S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  Lx, Ly, Lz: Single;
  vecLightDirUnnormalized: TD3DXVector3;
  light: TD3DLight9;

  vLight: TD3DXVector3;
  m: TD3DXMatrixA16;
begin
  // Setup the world spin matrix
  D3DXMatrixTranslation(m_matTerrainMatrix, 0.0, 0.0, 0.0);

  m_Camera.FrameMove(m_fElapsedTime);

  // Setup viewing postion from ArcBall
  m_matObjectMatrix := m_Camera.GetWorldMatrix^;

  m_pd3dDevice.SetTransform(D3DTS_VIEW,       m_Camera.GetViewMatrix^);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_Camera.GetProjMatrix^);

  // Move the light
  Lx :=   5;
  Ly :=   5;
  Lz :=  -5;
  vecLightDirUnnormalized := D3DXVector3(Lx, Ly, Lz);
  ZeroMemory(@light, SizeOf(TD3DLight9));
  light._Type        := D3DLIGHT_POINT;
  light.Diffuse.r   := 1.0;
  light.Diffuse.g   := 1.0;
  light.Diffuse.b   := 1.0;
  D3DXVec3Normalize(light.Direction, vecLightDirUnnormalized);
  light.Position.x   := Lx;
  light.Position.y   := Ly;
  light.Position.z   := Lz;
  light.Range        := 1000.0;
  light.Attenuation0 := 0.9;
  light.Attenuation1 := 0.0;
  m_pd3dDevice.SetLight(0, light);

  // Transform the light vector to be in object space
  D3DXMatrixInverse(m, nil, m_matObjectMatrix);
  vLight.x := Lx*m._11 + Ly*m._21 + Lz*m._31 + m._41;
  vLight.y := Lx*m._12 + Ly*m._22 + Lz*m._32 + m._42;
  vLight.z := Lx*m._13 + Ly*m._23 + Lz*m._33 + m._43;

  // Build the shadow volume
  m_pShadowVolume.Reset;
  m_pShadowVolume.BuildFromMesh(m_pAirplane.GetSysMemMesh, vLight);

  Result := S_OK;
end;


//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
begin
  // Clear the viewport, zbuffer, and stencil buffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER or D3DCLEAR_STENCIL,
                     $ff00bfff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matTerrainMatrix);
    m_pTerrainObject.Render(m_pd3dDevice);

    m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matObjectMatrix);
    m_pAirplane.Render(m_pd3dDevice);

    // Render the shadow volume into the stencil buffer, then add it into
    // the scene
    RenderShadow;
    DrawShadow;

    // Output statistics
    RenderText;

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result := S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RenderShadow()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.RenderShadow: HResult;
begin
  // Disable z-buffer writes (note: z-testing still occurs), and enable the
  // stencil-buffer
  m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE,  iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILENABLE, iTrue);

  // Dont bother with interpolating color
  m_pd3dDevice.SetRenderState(D3DRS_SHADEMODE,     D3DSHADE_FLAT);

  // Set up stencil compare fuction, reference value, and masks.
  // Stencil test passes if ((ref  and mask) cmpfn (stencil  and mask)) is true.
  // Note: since we set up the stencil-test to always pass, the STENCILFAIL
  // renderstate is really not needed.
  m_pd3dDevice.SetRenderState(D3DRS_STENCILFUNC,  D3DCMP_ALWAYS);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_KEEP);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILFAIL,  D3DSTENCILOP_KEEP);

  // If ztest passes, inc/decrement stencil buffer value
  m_pd3dDevice.SetRenderState(D3DRS_STENCILREF,       $1);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILMASK,      $ffffffff);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILWRITEMASK, $ffffffff);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILPASS,      D3DSTENCILOP_INCR);

  // Make sure that no pixels get drawn to the frame buffer
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,  D3DBLEND_ZERO);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);

  if ((m_d3dCaps.StencilCaps and D3DSTENCILCAPS_TWOSIDED) <> 0) then
  begin
    // With 2-sided stencil, we can avoid rendering twice:
    m_pd3dDevice.SetRenderState(D3DRS_TWOSIDEDSTENCILMODE, iTrue);
    m_pd3dDevice.SetRenderState(D3DRS_CCW_STENCILFUNC,  D3DCMP_ALWAYS);
    m_pd3dDevice.SetRenderState(D3DRS_CCW_STENCILZFAIL, D3DSTENCILOP_KEEP);
    m_pd3dDevice.SetRenderState(D3DRS_CCW_STENCILFAIL,  D3DSTENCILOP_KEEP);
    m_pd3dDevice.SetRenderState(D3DRS_CCW_STENCILPASS, D3DSTENCILOP_DECR);

    m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);

    // Draw both sides of shadow volume in stencil/z only
    m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matObjectMatrix);
    m_pShadowVolume.Render(m_pd3dDevice);

    m_pd3dDevice.SetRenderState(D3DRS_TWOSIDEDSTENCILMODE, iFalse);
  end else
  begin
    // Draw front-side of shadow volume in stencil/z only
    m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matObjectMatrix);
    m_pShadowVolume.Render(m_pd3dDevice);

    // Now reverse cull order so back sides of shadow volume are written.
    m_pd3dDevice.SetRenderState(D3DRS_CULLMODE, D3DCULL_CW);

    // Decrement stencil buffer value
    m_pd3dDevice.SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_DECR);

    // Draw back-side of shadow volume in stencil/z only
    m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matObjectMatrix);
    m_pShadowVolume.Render(m_pd3dDevice);
  end;

  // Restore render states
  m_pd3dDevice.SetRenderState(D3DRS_SHADEMODE, D3DSHADE_GOURAUD);
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE,  D3DCULL_CCW);
  m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE,     iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILENABLE,    iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);

  Result := S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DrawShadow()
// Desc: Draws a big gray polygon over scene according to the mask in the
//       stencil buffer. (Any pixel with stencil==1 is in the shadow.)
//-----------------------------------------------------------------------------
function CMyD3DApplication.DrawShadow: HResult;
begin
  // Set renderstates (disable z-buffering, enable stencil, disable fog, and
  // turn on alphablending)
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,          iFALSE);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILENABLE,    iTRUE);
  m_pd3dDevice.SetRenderState(D3DRS_FOGENABLE,        iFALSE);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTRUE);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,  D3DBLEND_SRCALPHA);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);

  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_MODULATE);

  // Only write where stencil val >= 1 (count indicates # of shadows that
  // overlap that pixel)
  m_pd3dDevice.SetRenderState(D3DRS_STENCILREF,  $1);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILFUNC, D3DCMP_LESSEQUAL);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILPASS, D3DSTENCILOP_KEEP);

  // Draw a big, gray square
  m_pd3dDevice.SetFVF(TShadowVertex_FVF);
  m_pd3dDevice.SetStreamSource(0, m_pBigSquareVB, 0, SizeOf(TShadowVertex));
  m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

  // Restore render states
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,          iTRUE );
  m_pd3dDevice.SetRenderState(D3DRS_STENCILENABLE,    iFALSE );
  m_pd3dDevice.SetRenderState(D3DRS_FOGENABLE,        iTRUE );
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFALSE );

  Result := S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RenderText()
// Desc: Draw the help & statistics for running sample
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.RenderText;
var
  rc: TRect;
begin
  if (nil = m_pFont) then Exit;

  // Output statistics
  SetRect(rc, 2, 0, 0, 0);
  m_pFont.DrawTextA(nil, m_strFrameStats, -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 0.0, 1.0)));
  SetRect(rc, 2, 15, 0, 0);
  m_pFont.DrawTextA(nil, m_strDeviceStats, -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 0.0, 1.0)));

  // Draw help
  if m_bShowHelp then
  begin
    SetRect(rc, 2, 30, 0, 0);
    m_pFont.DrawTextA(nil, 'Controls:', -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));
    SetRect(rc, 20, 45, 0, 0);
    m_pFont.DrawTextA(nil, 'Pitch/yaw model'#10 +
                           'Roll model'#10 +
                           'Rotate camera'#10 +
                           'Zoom camera'#10 +
                           'Move'#10'Move up/down'#10 +
                           'Reset camera'#10 +
                           'Show controls'#10'Change Device'#10'Quit',
                           -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));
    SetRect(rc, 240, 45, 0, 0);
    m_pFont.DrawTextA(nil, 'Left mouse button'#10 +
                           'Middle mouse button'#10 +
                           'Right mouse button'#10 +
                           'Mouse wheel scroll'#10 +
                           'A/W/S/D or Arrow Keys'#10'Q/E or PgUp/PgDn'#10 +
                           'Home'#10 +
                           'F1'#10'F2'#10'Esc',
                           -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));
  end else
  begin
    SetRect(rc, 2, 30, 0, 0);
    m_pFont.DrawTextA(nil, 'Press F1 for help.', -1, @rc, DT_NOCLIP, D3DXColorToDWord(D3DXColor(1.0, 1.0, 1.0, 1.0)));
  end;
end;



//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  if Assigned(m_pFont) then m_pFont.OnLostDevice;
  m_pAirplane.InvalidateDeviceObjects;
  m_pTerrainObject.InvalidateDeviceObjects;
  Result := S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
begin
  m_pFont := nil;
  m_pAirplane.Destroy_;
  m_pTerrainObject.Destroy_;

  m_pBigSquareVB:= nil;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  FreeAndNil(m_pAirplane);
  FreeAndNil(m_pTerrainObject);
  FreeAndNil(m_pShadowVolume);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9;
  dwBehavior: DWORD; adapterFormat, backBufferFormat: TD3DFormat): HRESULT;
begin
  Result := E_FAIL;

  // Make sure device supports point lights
  if ((dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING) or
      (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING)) <> 0 then
  begin
    if (0 = (pCaps.VertexProcessingCaps and D3DVTXPCAPS_POSITIONALLIGHTS))
    then Exit;

    // Need to support post-pixel processing (for fog and stencil)
    if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
        adapterFormat, D3DUSAGE_RENDERTARGET or D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING,
        D3DRTYPE_SURFACE, backBufferFormat))
    then Exit;
  end;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT;
begin
  // Pass mouse messages to the ArcBall so it can build internal matrices
  m_Camera.HandleMessages(hWnd, uMsg, wParam, lParam);

  case uMsg of
    WM_COMMAND:
    begin
      case LOWORD(wParam) of
        IDM_TOGGLEHELP: m_bShowHelp := not m_bShowHelp;
      end;
    end;
  end;

  // Trap the context menu
  if (WM_CONTEXTMENU = uMsg) then
  begin
    Result := 0;
    Exit;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.

