(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: StencilDepth.cpp
//
// Desc: Example code showing how to use stencil buffers to show the depth
//       complexity of a scene.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit StencilDepthUnit;

{$I DirectX.inc}

interface

uses
  Windows, Messages,  
  Direct3D9, D3DX9,
  D3DApp, D3DFont, D3DFile, D3DUtil, DXUtil,
  Resource;

type

  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont: CD3DFont;
    m_bShowDepthComplexity: Boolean;
    m_pFileObject: CD3DFile;

    m_pBigSquareVB: IDirect3DVertexBuffer9;
    m_matWorldMatrix: TD3DXMatrixA16;

    procedure SetStatesForRecordingDepthComplexity;
    procedure ShowDepthComplexity;

  protected
    function InitDeviceObjects: HRESULT; override;
    function RestoreDeviceObjects: HRESULT; override;
    function InvalidateDeviceObjects: HRESULT; override;
    function DeleteDeviceObjects: HRESULT; override;
    function Render: HRESULT; override;
    function FrameMove: HRESULT; override;
    function FinalCleanup: HRESULT; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;

  public
    constructor Create;

    function MsgProc(hWnd: HWND; uMsg: LongWord; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  end;

implementation



{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;
  m_strWindowTitle := 'StencilDepth: Displaying Depth Complexity';
  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_d3dEnumeration.AppMinDepthBits := 16;
  m_d3dEnumeration.AppMinStencilBits := 4;

  m_bShowDepthComplexity := True;
  m_pFont         := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pFileObject   := CD3DFile.Create;
  m_pBigSquareVB  := nil;

  D3DXMatrixIdentity(m_matWorldMatrix);
end;


//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HRESULT;
var
  vAxis: TD3DXVector3;
begin
  // Setup the world spin matrix
  vAxis:= D3DXVector3(1.0, 1.0, 0.0);
  D3DXMatrixRotationAxis(m_matWorldMatrix, vAxis, m_fTime/2);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: SetStatesForRecordingDepthComplexity()
// Desc: Turns on stencil and other states for recording the depth complexity
//       during the rendering of a scene.
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.SetStatesForRecordingDepthComplexity;
begin
  // Clear the stencil buffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_STENCIL, $0, 1.0, 0);

  // Turn stenciling
  m_pd3dDevice.SetRenderState(D3DRS_STENCILENABLE,   iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILFUNC,     D3DCMP_ALWAYS);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILREF,      0);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILMASK,     $00000000);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILWRITEMASK,$ffffffff);

  // Increment the stencil buffer for each pixel drawn
  m_pd3dDevice.SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_INCRSAT);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILFAIL,  D3DSTENCILOP_KEEP);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILPASS,  D3DSTENCILOP_INCRSAT);
end;


//-----------------------------------------------------------------------------
// Name: ShowDepthComplexity()
// Desc: Draws the contents of the stencil buffer in false color. Use alpha
//       blending of one red, one green, and one blue rectangle to do false
//       coloring of bits 1, 2, and 4 in the stencil buffer.
//-----------------------------------------------------------------------------
procedure CMyD3DApplication.ShowDepthComplexity;
begin
  // Turn off the buffer, and enable alpha blending
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,          iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,  D3DBLEND_SRCCOLOR);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCCOLOR);

  // Set up the stencil states
  m_pd3dDevice.SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_KEEP);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILFAIL,  D3DSTENCILOP_KEEP);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILPASS,  D3DSTENCILOP_KEEP);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILFUNC,  D3DCMP_NOTEQUAL);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILREF,   0);

  // Set the background to black
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET, $00000000, 1.0, 0);

  // Set render states for drawing a rectangle that covers the viewport.
  // The color of the rectangle will be passed in D3DRS_TEXTUREFACTOR
  m_pd3dDevice.SetFVF(D3DFVF_XYZRHW);
  m_pd3dDevice.SetStreamSource(0, m_pBigSquareVB, 0, SizeOf(TD3DXVector4));
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TFACTOR);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_SELECTARG1);

  // Draw a red rectangle wherever the 1st stencil bit is set
  m_pd3dDevice.SetRenderState(D3DRS_STENCILMASK, $01);
  m_pd3dDevice.SetRenderState(D3DRS_TEXTUREFACTOR, $ffff0000);
  m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

  // Draw a green rectangle wherever the 2nd stencil bit is set
  m_pd3dDevice.SetRenderState(D3DRS_STENCILMASK, $02);
  m_pd3dDevice.SetRenderState(D3DRS_TEXTUREFACTOR, $ff00ff00);
  m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

  // Draw a blue rectangle wherever the 3rd stencil bit is set
  m_pd3dDevice.SetRenderState(D3DRS_STENCILMASK, $04);
  m_pd3dDevice.SetRenderState(D3DRS_TEXTUREFACTOR, $ff0000ff);
  m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

  // Restore states
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,          iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILENABLE,    iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);
end;


//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HRESULT;
begin
  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER, $000000ff, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    if m_bShowDepthComplexity then SetStatesForRecordingDepthComplexity;

    // Render the scene
    m_pFileObject.Render(m_pd3dDevice, @m_matWorldMatrix);

    // Show the depth complexity of the scene
    if m_bShowDepthComplexity then ShowDepthComplexity;

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HRESULT;
begin
  // Initialize the font's internal textures
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  // Load a .X file
  if FAILED(m_pFileObject.CreateFromFile(m_pd3dDevice, 'Heli.x')) then
  begin
    Result:= D3DAPPERR_MEDIANOTFOUND;
    Exit;
  end;

  // Create a big square for rendering the stencilbuffer contents
  if FAILED(m_pd3dDevice.CreateVertexBuffer(4*SizeOf(TD3DXVector4),
              D3DUSAGE_WRITEONLY, D3DFVF_XYZRHW,
              D3DPOOL_MANAGED, m_pBigSquareVB, nil)) then
  begin
    Result:= E_FAIL;
    Exit;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HRESULT;
type
  PD3DXVector4Array = ^TD3DXVector4Array;
  TD3DXVector4Array = array [0..MaxInt div SizeOf(TD3DXVector4) - 1] of TD3DXVector4; 
var
  vEyePt: TD3DXVector3;
  vLookatPt: TD3DXVector3;
  vUpVec: TD3DXVector3;
  {matWorld, }matView, matProj: TD3DXMatrixA16;
  fAspect: Single;
  mtrl: TD3DMaterial9;
  light: TD3DLight9;

  v: PD3DXVector4Array;
  sx: Single;
  sy: Single;
begin
  m_pFont.RestoreDeviceObjects;

  Result:= m_pFileObject.RestoreDeviceObjects(m_pd3dDevice);
  if FAILED(Result) then Exit;

  // Setup textures (the .X file may have textures)
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE,   iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,        iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_COLORVERTEX,    iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,        $00000000);

  // Set the transform matrices
  vEyePt    := D3DXVector3(0.0, 0.0,-15.0);
  vLookatPt := D3DXVector3(0.0, 0.0,  0.0);
  vUpVec    := D3DXVector3(0.0, 1.0,  0.0);

  D3DXMatrixIdentity(m_matWorldMatrix);
  D3DXMatrixLookAtLH(matView, vEyePt, vLookatPt, vUpVec);
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 100.0);

  m_pd3dDevice.SetTransform(D3DTS_VIEW,       matView);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Setup a material
  D3DUtil_InitMaterial(mtrl, 1.0, 1.0, 1.0);
  m_pd3dDevice.SetMaterial(mtrl);

  D3DUtil_InitLight(light, D3DLIGHT_DIRECTIONAL, 0.0, -1.0, 0.0);
  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, True);
  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTrue);

  // Resize the quad used to render the stencilbuffer contents
  sx := m_d3dsdBackBuffer.Width;
  sy := m_d3dsdBackBuffer.Height;
  m_pBigSquareVB.Lock(0, 0, Pointer(v), 0);
  v[0] := D3DXVector4( 0, sy, 0.0, 1.0);
  v[1] := D3DXVector4( 0,  0, 0.0, 1.0);
  v[2] := D3DXVector4(sx, sy, 0.0, 1.0);
  v[3] := D3DXVector4(sx,  0, 0.0, 1.0);
  m_pBigSquareVB.Unlock;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HRESULT;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pFileObject.InvalidateDeviceObjects;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HRESULT;
begin
  m_pFont.DeleteDeviceObjects;
  m_pFileObject.Destroy_;

  SAFE_RELEASE(m_pBigSquareVB);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HRESULT;
begin
  SAFE_DELETE(m_pFont);
  SAFE_DELETE(m_pFileObject);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9;
  dwBehavior: DWORD; adapterFormat, backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  // Make sure device supports directional lights
  if ((dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING <> 0) or
      (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING <> 0)) then
  begin
    if (0 = ( pCaps.VertexProcessingCaps and D3DVTXPCAPS_DIRECTIONALLIGHTS)) then Exit;
  end;

  // Need to support post-pixel processing (for stencil)
  if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
      adapterFormat, D3DUSAGE_RENDERTARGET or D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING,
      D3DRTYPE_SURFACE, backBufferFormat))
  then Exit;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Overrrides the main WndProc, so the sample can do custom message
//       handling (e.g. processing mouse, keyboard, or menu commands).
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: LongWord;
  wParam: WPARAM; lParam: LPARAM): LRESULT;
const
  IsChecked: array[False..True] of Integer = (MF_UNCHECKED, MF_CHECKED);
begin
  if (WM_COMMAND = uMsg) then
  begin
    case LOWORD(wParam) of
      IDM_SHOWDEPTHCOMPLEXITY:
      begin
        m_bShowDepthComplexity := not m_bShowDepthComplexity;
        CheckMenuItem(GetMenu(hWnd), IDM_SHOWDEPTHCOMPLEXITY,
          IsChecked[m_bShowDepthComplexity]);
      end;
    end;
  end;

  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

end.
