(*----------------------------------------------------------------------------*
 *  Direct3D sample from DirectX 9.0 SDK Summer 2003                          *
 *  Delphi adaptation by Alexey Barkovoy (e-mail: clootie@ixbt.com)           *
 *                                                                            *
 *  Delphi versions 5-7 are supported                                         *
 *                                                                            *
 *  Modified: 30-Dec-2003                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru/delphi                                         *
 *----------------------------------------------------------------------------*)
//-----------------------------------------------------------------------------
// File: StencilMirror.cpp
//
// Desc: Example code showing how to use stencil buffers to implement planar
//       mirrors.
//
//       Note: This code uses the D3D Framework helper library.
//
// Copyright (c) Microsoft Corporation. All rights reserved.
//-----------------------------------------------------------------------------

unit StencilMirrorUnit;

interface

uses
  Windows, Messages, SysUtils,
  Direct3D9, D3DX9,
  D3DApp, D3DFile, D3DFont, D3DUtil, DXUtil;

type
  //-----------------------------------------------------------------------------
  // Custom vertex types
  //-----------------------------------------------------------------------------
  TMeshVertex = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
    tu, tv: Single;
  end;
  PMeshVertexArray = ^TMeshVertexArray;
  TMeshVertexArray = array [0..0] of TMeshVertex;

const
  TMeshVertex_FVF = (D3DFVF_XYZ or D3DFVF_NORMAL or D3DFVF_TEX1);

type
  TMirrorVertex = packed record
    p: TD3DXVector3;
    n: TD3DXVector3;
  end;
  PMirrorVertexArray = ^TMirrorVertexArray;
  TMirrorVertexArray = array [0..10000] of TMirrorVertex;

const
  TMirrorVertex_FVF = (D3DFVF_XYZ or D3DFVF_NORMAL);

  FOG_COLOR = $00000080;

function FtoDW(f: Single): DWORD;

function HeightField(x, z: Single): Single;

type
  //-----------------------------------------------------------------------------
  // Name: class CMyD3DApplication
  // Desc: Application class. The base class (CD3DApplication) provides the
  //       generic functionality needed in all Direct3D samples. CMyD3DApplication
  //       adds functionality specific to this sample program.
  //-----------------------------------------------------------------------------
  CMyD3DApplication = class(CD3DApplication)
  private
    m_pFont:            CD3DFont;


    m_pMirrorVB: IDirect3DVertexBuffer9;
    m_mtrlMirrorMaterial:  TD3DMaterial9;    // Material of the mirror
    m_matMirrorMatrix:     TD3DXMatrixA16;   // Matrix to position mirror

    m_pTerrain:            CD3DFile;         // X file of terrain
    m_matTerrainMatrix:    TD3DXMatrixA16;   // Matrix to position terrain

    m_pHelicopter:         CD3DFile;         // X file object to render
    m_matHelicopterMatrix: TD3DXMatrix;      // Matrix to animate X file object

    function RenderScene: HResult;
    function RenderMirror: HResult;

  protected
    function OneTimeSceneInit: HResult; override;
    function InitDeviceObjects: HResult; override;
    function RestoreDeviceObjects: HResult; override;
    function InvalidateDeviceObjects: HResult; override;
    function DeleteDeviceObjects: HResult; override;
    function Render: HResult; overload; override;
    function FrameMove: HResult; override;
    function FinalCleanup: HResult; override;
  public
    constructor Create;

    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
 end;

implementation

function FtoDW(f: Single): DWORD;
begin
  Result:= PDWord(@f)^;
end;

function HeightField(x, z: Single): Single;
var
  y: Single;
begin
  y := 0.0;
  y:= y + 7.0 * Cos(0.051*x + 0.0) * Sin(0.055*x + 0.0);
  y:= y + 7.0 * Cos(0.053*z + 0.0) * Sin(0.057*z + 0.0);
  y:= y + 1.0 * Cos(0.101*x + 0.0) * Sin(0.105*x + 0.0);
  y:= y + 1.0 * Cos(0.103*z + 0.0) * Sin(0.107*z + 0.0);
  y:= y + 1.0 * Cos(0.251*x + 0.0) * Sin(0.255*x + 0.0);
  y:= y + 1.0 * Cos(0.253*z + 0.0) * Sin(0.257*z + 0.0);
  Result := y;
end;


{ CMyD3DApplication }

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited Create;

  m_strWindowTitle    := 'StencilMirror: Doing Reflections with Stencils';
  m_d3dEnumeration.AppUsesDepthBuffer := True;
  m_d3dEnumeration.AppMinDepthBits    := 16;
  m_d3dEnumeration.AppMinStencilBits  := 4;

  m_pFont             := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_pTerrain          := CD3DFile.Create;
  m_pHelicopter       := CD3DFile.Create;
  m_pMirrorVB         := nil;
end;

//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HResult;
begin
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HResult;
var
  fObjectPosX, fObjectPosY, fObjectPosZ: Single;
  matRoll, matPitch, matRotate, matScale, matTranslate: TD3DXMatrixA16;

  fEyeX, fEyeY, fEyeZ: Single;

  matView: TD3DXMatrixA16;
  vEyePt, vLookatPt, vUpVec: TD3DXVector3;
begin
  // Position the terrain
  D3DXMatrixTranslation(m_matTerrainMatrix, 0.0, 8.0, 0.0);

  // Position the mirror (water polygon intersecting with terrain)
  D3DXMatrixTranslation(m_matMirrorMatrix, 0.0, 0.0, 0.0);

  // Position and animate the main object
  fObjectPosX := 50.0*Sin(m_fTime/2);
  fObjectPosY := 6;
  fObjectPosZ := 10.0*Cos(m_fTime/2);
  D3DXMatrixRotationZ(matRoll, 0.2*Sin(m_fTime/2));
  D3DXMatrixRotationY(matRotate, m_fTime/2-D3DX_PI/2);
  D3DXMatrixRotationX(matPitch, -0.1 * (1+Cos(m_fTime)));
  D3DXMatrixScaling(matScale, 0.5, 0.5, 0.5);
  D3DXMatrixTranslation(matTranslate, fObjectPosX, fObjectPosY, fObjectPosZ);
  D3DXMatrixMultiply(m_matHelicopterMatrix, matScale,  matTranslate);
  D3DXMatrixMultiply(m_matHelicopterMatrix, matRoll,   m_matHelicopterMatrix);
  D3DXMatrixMultiply(m_matHelicopterMatrix, matRotate, m_matHelicopterMatrix);
  D3DXMatrixMultiply(m_matHelicopterMatrix, matPitch,  m_matHelicopterMatrix);

  // Move the camera around
  fEyeX := 10.0 * Sin( m_fTime/2.0);
  fEyeY :=  3.0 * Sin( m_fTime/25.0) + 13.0;
  fEyeZ :=  5.0 * Cos( m_fTime/2.0);
  vEyePt    := D3DXVector3( fEyeX, fEyeY, fEyeZ);
  vLookatPt := D3DXVector3( fObjectPosX, fObjectPosY, fObjectPosZ);
  vUpVec    := D3DXVector3( 0.0, 1.0, 0.0);
  D3DXMatrixLookAtLH(matView, vEyePt, vLookatPt, vUpVec);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, matView);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RenderScene()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.RenderScene: HResult;
begin
  // Render terrain
  m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matTerrainMatrix);
  m_pTerrain.Render(m_pd3dDevice);

  // Draw the mirror
  m_pd3dDevice.SetTexture(0, nil);
  m_pd3dDevice.SetMaterial(m_mtrlMirrorMaterial);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matMirrorMatrix);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,     D3DBLEND_DESTCOLOR);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND,    D3DBLEND_ZERO);
  m_pd3dDevice.SetFVF(TMirrorVertex_FVF);
  m_pd3dDevice.SetStreamSource(0, m_pMirrorVB, 0, SizeOf(TMirrorVertex));
  m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);

  // Draw the object. Note: do this last, in  the object has alpha
  m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matHelicopterMatrix);
  m_pHelicopter.Render(m_pd3dDevice);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RenderMirror()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.RenderMirror: HResult;
var
  matViewSaved: TD3DXMatrix;

  matView, matReflect: TD3DXMatrix;
  plane: TD3DXPlane;

begin
  // Turn depth buffer off, and stencil buffer on
  m_pd3dDevice.SetRenderState(D3DRS_STENCILENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILFUNC,     D3DCMP_ALWAYS);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILREF,      $1);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILMASK,     $ffffffff);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILWRITEMASK,$ffffffff);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILZFAIL, D3DSTENCILOP_KEEP);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILFAIL,  D3DSTENCILOP_KEEP);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILPASS,  D3DSTENCILOP_REPLACE);

  // Make sure no pixels are written to the z-buffer or frame buffer
  m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,  D3DBLEND_ZERO);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND, D3DBLEND_ONE);

  // Draw the reflecting surface into the stencil buffer
  m_pd3dDevice.SetTexture(0, nil);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matMirrorMatrix);
  m_pd3dDevice.SetFVF(TMirrorVertex_FVF);
  m_pd3dDevice.SetStreamSource(0, m_pMirrorVB, 0, SizeOf(TMirrorVertex));
  m_pd3dDevice.DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

  // Save the view matrix
  m_pd3dDevice.GetTransform(D3DTS_VIEW, matViewSaved);

  // Reflect camera in X-Z plane mirror
  D3DXPlaneFromPointNormal(plane, D3DXVector3(0,0,0), D3DXVector3(0,1,0));
  D3DXMatrixReflect(matReflect, plane);
  D3DXMatrixMultiply(matView, matReflect, matViewSaved);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, matView);

  // Set a clip plane, so that only objects above the water are reflected
  m_pd3dDevice.SetClipPlane(0, @plane.a);
  m_pd3dDevice.SetRenderState(D3DRS_CLIPPLANEENABLE, $01);

  // Setup render states to a blended render scene against mask in stencil
  // buffer. An important step here is to reverse the cull-order of the
  // polygons, since the view matrix is being relected.
  m_pd3dDevice.SetRenderState(D3DRS_ZWRITEENABLE, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILFUNC,  D3DCMP_EQUAL);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILPASS,  D3DSTENCILOP_KEEP);
  m_pd3dDevice.SetRenderState(D3DRS_SRCBLEND,     D3DBLEND_DESTCOLOR);
  m_pd3dDevice.SetRenderState(D3DRS_DESTBLEND,    D3DBLEND_ZERO);
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE,     D3DCULL_CW);

  // Clear the zbuffer (leave frame- and stencil-buffer intact)
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_ZBUFFER, 0, 1.0, 0);

  // Render the scene
  RenderScene;

  // Restore render states
  m_pd3dDevice.SetRenderState(D3DRS_CULLMODE,         D3DCULL_CCW);
  m_pd3dDevice.SetRenderState(D3DRS_STENCILENABLE,    iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_ALPHABLENDENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_CLIPPLANEENABLE,  $00);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, matViewSaved);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HResult;
begin
  // Clear the viewport, zbuffer, and stencil buffer
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER or D3DCLEAR_STENCIL,
                     FOG_COLOR, 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    // Render the scene
    RenderScene;

    // Render the reflection in the mirror
    RenderMirror;

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    // End the scene.
    m_pd3dDevice.EndScene;
  end;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects: HResult;
var
  pMesh: CD3DMesh;

  pVB: IDirect3DVertexBuffer9;
  pVertices: PMeshVertexArray;
  dwNumVertices: DWORD;

  i: Integer;

  v00, v10, v01: TD3DXVector3;
  n: TD3DXVector3;
  v1, v2: TD3DXVector3;

  v: PMirrorVertexArray;
begin
  // Initialize the font's internal textures
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  Result := D3DAPPERR_MEDIANOTFOUND;

  // Load the main file object
  if FAILED(m_pHelicopter.CreateFromFile(m_pd3dDevice, 'Heli.x'))
  then Exit;

  // Load the terrain
  if FAILED(m_pTerrain.CreateFromFile(m_pd3dDevice, 'SeaFloor.x'))
  then Exit;

  // Tweak the terrain vertices to add some bumpy terrain
  pMesh := m_pTerrain.FindMesh('SeaFloor');
  if Assigned(pMesh) then
  begin
    // Set FVF to VertexFVF
    pMesh.SetFVF(m_pd3dDevice, TMeshVertex_FVF);

    // Get access to the mesh vertices
    dwNumVertices := pMesh.GetSysMemMesh.GetNumVertices;
    pMesh.GetSysMemMesh.GetVertexBuffer(pVB);
    pVB.Lock(0, 0, Pointer(pVertices), 0);

    for i:= 0 to (dwNumVertices - 1) do
    begin
      v00:= D3DXVector3(pVertices[i].p.x + 0.0, 0.0, pVertices[i].p.z + 0.0);
      v10:= D3DXVector3(pVertices[i].p.x + 0.1, 0.0, pVertices[i].p.z + 0.0);
      v01:= D3DXVector3(pVertices[i].p.x + 0.0, 0.0, pVertices[i].p.z + 0.1);
      v00.y := HeightField(1*v00.x, 1*v00.z);
      v10.y := HeightField(1*v10.x, 1*v10.z);
      v01.y := HeightField(1*v01.x, 1*v01.z);

      D3DXVec3Subtract(v1, v01, v00);
      D3DXVec3Subtract(v2, v10, v00);
      D3DXVec3Cross(n, v1, v2);
      D3DXVec3Normalize(n, n);

      pVertices[i].p.y  := v00.y;
      pVertices[i].n.x  := n.x;
      pVertices[i].n.y  := n.y;
      pVertices[i].n.z  := n.z;
      pVertices[i].tu   := pVertices[i].tu * 10;
      pVertices[i].tv   := pVertices[i].tv * 10;
    end;

    pVB.Unlock;
    pVB:= nil;
  end;

  // Create a big square for rendering the mirror
  if FAILED( m_pd3dDevice.CreateVertexBuffer( 4*SizeOf(TMirrorVertex),
      D3DUSAGE_WRITEONLY, TMirrorVertex_FVF, D3DPOOL_MANAGED, m_pMirrorVB, nil)) then
  begin
    Result := E_FAIL;
    Exit;
  end;

  m_pMirrorVB.Lock( 0, 0, Pointer(v), 0);
  v[0].p := D3DXVector3(-80.0, 0.0,-80.0);
  v[0].n := D3DXVector3( 0.0,  1.0,  0.0);
  v[1].p := D3DXVector3(-80.0, 0.0, 80.0);
  v[1].n := D3DXVector3( 0.0,  1.0,  0.0);
  v[2].p := D3DXVector3( 80.0, 0.0,-80.0);
  v[2].n := D3DXVector3( 0.0,  1.0,  0.0);
  v[3].p := D3DXVector3( 80.0, 0.0, 80.0);
  v[3].n := D3DXVector3( 0.0,  1.0,  0.0);
  m_pMirrorVB.Unlock;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HResult;
var
  vEyePt, vLookatPt, vUpVec: TD3DXVector3;
  matWorld, matView, matProj: TD3DXMatrixA16;

  fAspect: Single;
  vecLightDirUnnormalized: TD3DXVector3;
  light: TD3DLight9;
  fFogStart, fFogEnd: Single;

begin
  m_pFont.RestoreDeviceObjects;

  // Build the device objects for the file-based objecs
  m_pHelicopter.RestoreDeviceObjects(m_pd3dDevice);
  m_pTerrain.RestoreDeviceObjects(m_pd3dDevice);

  // Set up textures
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
  m_pd3dDevice.SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

  // Set up misc render states
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE,        iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_DITHERENABLE,   iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_SPECULARENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT,        $00555555);

  // Set the transform matrices
  vEyePt    := D3DXVector3( 0.0, 5.5, -15.0);
  vLookatPt := D3DXVector3( 0.0, 1.5,   0.0);
  vUpVec    := D3DXVector3( 0.0, 1.0,   0.0);

  D3DXMatrixIdentity(matWorld);
  D3DXMatrixLookAtLH(matView, vEyePt, vLookatPt, vUpVec);
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(matProj, D3DX_PI/4, fAspect, 1.0, 1000.0);

  m_pd3dDevice.SetTransform(D3DTS_WORLD,      matWorld);
  m_pd3dDevice.SetTransform(D3DTS_VIEW,       matView);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, matProj);

  // Set up a material
  ZeroMemory(@m_mtrlMirrorMaterial, SizeOf(TD3DMaterial9));
  m_mtrlMirrorMaterial.Diffuse.r := 0.5; m_mtrlMirrorMaterial.Ambient.r := 0.5;
  m_mtrlMirrorMaterial.Diffuse.g := 0.8; m_mtrlMirrorMaterial.Ambient.g := 0.8;
  m_mtrlMirrorMaterial.Diffuse.b := 1.0; m_mtrlMirrorMaterial.Ambient.b := 1.0;
  m_mtrlMirrorMaterial.Diffuse.a := 1.0; m_mtrlMirrorMaterial.Ambient.a := 1.0;

  // Set up the light
  vecLightDirUnnormalized := D3DXVector3(0.0, -1.0, 1.0);
  ZeroMemory(@light, SizeOf(TD3DLight9));
  light._Type       := D3DLIGHT_DIRECTIONAL;
  light.Diffuse.r   := 1.0;
  light.Diffuse.g   := 1.0;
  light.Diffuse.b   := 1.0;
  D3DXVec3Normalize(light.Direction, vecLightDirUnnormalized);
  light.Position.x   := 0.0;
  light.Position.y   := -1.0;
  light.Position.z   := 1.0;
  light.Range        := 1000.0;
  m_pd3dDevice.SetLight(0, light);
  m_pd3dDevice.LightEnable(0, True);

  // Turn on fog
  fFogStart :=  80.0;
  fFogEnd   := 100.0;
  m_pd3dDevice.SetRenderState(D3DRS_FOGENABLE,      iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_FOGCOLOR,       FOG_COLOR);
  m_pd3dDevice.SetRenderState(D3DRS_FOGTABLEMODE,   D3DFOG_NONE);
  m_pd3dDevice.SetRenderState(D3DRS_FOGVERTEXMODE,  D3DFOG_LINEAR);
  m_pd3dDevice.SetRenderState(D3DRS_RANGEFOGENABLE, iFalse);
  m_pd3dDevice.SetRenderState(D3DRS_FOGSTART,       FtoDW(fFogStart));
  m_pd3dDevice.SetRenderState(D3DRS_FOGEND,         FtoDW(fFogEnd));

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HResult;
begin
  m_pFont.InvalidateDeviceObjects;
  m_pTerrain.InvalidateDeviceObjects;
  m_pHelicopter.InvalidateDeviceObjects;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HResult;
begin
  m_pFont.DeleteDeviceObjects;
  m_pTerrain.Destroy_;
  m_pHelicopter.Destroy_;

  m_pMirrorVB:= nil;

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HResult;
begin
  FreeAndNil(m_pFont);
  FreeAndNil(m_pTerrain);
  FreeAndNil(m_pHelicopter);

  Result := S_OK;
end;

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9;
  dwBehavior: DWORD; adapterFormat, backBufferFormat: TD3DFormat): HRESULT;
begin
  Result := E_FAIL;

  // GetTransform doesn't work on PUREDEVICE
  if (dwBehavior and D3DCREATE_PUREDEVICE) <> 0 then Exit;

  // Make sure device supports directional lights
  if ((dwBehavior and D3DCREATE_HARDWARE_VERTEXPROCESSING <> 0) or
      (dwBehavior and D3DCREATE_MIXED_VERTEXPROCESSING <> 0)) then
  begin
    if (0 = (pCaps.VertexProcessingCaps and D3DVTXPCAPS_DIRECTIONALLIGHTS)) then Exit;
    if (pCaps.MaxUserClipPlanes < 1) then Exit;
  end;

  // Need to support post-pixel processing (for fog and stencil)
  if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
      adapterFormat, D3DUSAGE_RENDERTARGET or D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING,
      D3DRTYPE_SURFACE, backBufferFormat))
  then Exit;

  Result:= S_OK;
end;

end.

